#include <stdio.h>
#include <unistd.h>
#include <strings.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <string.h>
#include <fcntl.h>
#include "disco.h"

/* a very simple TFTP server for receiving files only */

#define TFTPDPORT 69

#define MODE_NONE 0
#define MODE_NETASCII 1
#define MODE_OTHER 2

int tftp_setup() {
  /* returns -1 on error, otherwise returns the socket fd */
  int sockfd, ret;
  struct sockaddr_in servaddr;

  /* create the socket */
  sockfd=socket(AF_INET, SOCK_DGRAM, 0);
  if (sockfd<0) {
    perror("Error creating socket");
    return(-1);
  }
  if (DEBUG) printf("Socket created\n");

  /* set it non blocking */
  ret=fcntl(sockfd, F_GETFL, 0);
  fcntl(sockfd, F_SETFL, ret|O_NONBLOCK);

  /* server address */
  memset(&servaddr, 0, sizeof(servaddr));
  servaddr.sin_family=AF_INET;
  servaddr.sin_addr.s_addr=htonl(INADDR_ANY);
  servaddr.sin_port=htons(TFTPDPORT);

  /* bind the address */
  ret=bind(sockfd, (struct sockaddr *) &servaddr, sizeof(servaddr));
  if (ret) {
    perror("Error binding socket");
    return(-1);
  }
  if (DEBUG) printf("Socket bound\n");

  return(sockfd);
}

int tftp_server(int sockfd, char *filename, char *hostname) {
  char *ptr;
  int len, packetlen, mode, ret, t, retrans;
  int opcode, block, curblock, origport;
  char rbuff[1500], sbuff[1500];
  struct sockaddr_in origaddr, cliaddr;
  FILE *file;
  
  /* open file for writing */
  file=fopen(filename, "w");
  if (!file) {
    perror("Could not open file for writing");
    return(-1);
  }
  if (verbose) printf("Writing to %s\n", filename);

  /* clear the client address */
  memset(&cliaddr, 0, sizeof(cliaddr));

  /* main loop */
  origport=0;
  curblock=0;
  mode=MODE_NONE;
  retrans=0;
  t=(int) time(NULL);
  if (verbose) printf("Waiting for initial TFTP packet from router...\n");
  
  while (1) {
    usleep(3);

    if (((int) time(NULL)-t) >= 10) {
      printf("No packet received for 10 seconds.  Timing out.\n");
      return(1);
    }

    if (!retrans && ((int) time(NULL)-t) >= 5) {
      if (verbose) printf("No packet received for 5 seconds; retransmitting ...\n");
      sendto(sockfd, sbuff, 4, NULL, (struct sockaddr *) &cliaddr, sizeof(cliaddr));
      retrans=1;
    }

    len=sizeof(cliaddr);
    packetlen=recvfrom(sockfd, rbuff, 1500, 0, (struct sockaddr *) &cliaddr, &len);
    if (packetlen<=0) {
      continue;
    }

    t=time(NULL);
    retrans=0;
    
    if (verbose && mode==MODE_NONE) {
      printf("Initial packet received from %s:%i.\n",inet_ntoa(cliaddr.sin_addr), ntohs(cliaddr.sin_port));
    }

    opcode=(((unsigned char) rbuff[0])*256) + (unsigned char) rbuff[1];
    
    /* first packet should be the write request */
    if (mode==MODE_NONE) {

      /* save the port and cliaddr */
      origport=ntohs(cliaddr.sin_port);
      memcpy(&origaddr, &cliaddr, sizeof(cliaddr));

      /* check that it's a write request */
      if (opcode==2) {
	/* send an ack */
	make_ack(sbuff, 0);
	sendto(sockfd, sbuff, 4, NULL, (struct sockaddr *) &cliaddr, sizeof(cliaddr));
	if (DEBUG) printf("Got write req, sent ack\n");

	/* get the mode */
	ptr=strchr(rbuff+2, 0);
	if (!ptr) {
	  printf("Could not find mode in initial tftp request\n");
	  return(-1);
	}
	if (!strcmp(ptr+1, "netascii")) {
	  mode=MODE_NETASCII;
	  if (verbose) printf("Mode is netascii.\n");
	  if (DEBUG) printf("Have mode netascii.\n");
	} else {
	  mode=MODE_OTHER;
	  if (verbose) printf("Mode is raw.\n");
	}
	continue;
      } else if (opcode==1) {
	printf("Received tftp opcode for read, only write accepted\n");
	return(-1);
      } else if (opcode==5) {
	printf("tftp error packet received\n");
	return(-1);
      } else {
	printf("Strange tftp opcode %i received before initial request\n", opcode);
	return(-1);
      }
    }

    /* check that we're coming from the same host/port and we started with */
    if (ntohs(cliaddr.sin_port) != origport) {
      if (DEBUG) printf("Received packet from different port (%i) vs (%i)\n",
			ntohs(cliaddr.sin_port),
			origport);
      continue;
    }
    if (memcmp(&cliaddr.sin_addr.s_addr, &origaddr.sin_addr.s_addr, 4)) {
      if (DEBUG) printf("Received packet from different address (%s) vs (%s)\n",
			inet_ntoa(cliaddr.sin_addr),
			inet_ntoa(origaddr.sin_addr));
      continue;
    }

    /* bail if we get an error */
    if (opcode==5) {
      printf("Received tftp error packet\n");
      return(-1);
    }
    /* bail if we get any other than a data packet */
    if (opcode!=3) {
      printf("Unexpected tftp opcode %i received\n", opcode);
      return(-1);
    }

    /* See if we're getting the next contiguous block.  If not we'll
       just be a bastard and not ack it in the hopes that we'll get
       it again later when we want it.  Really robust code wouldn't
       do this, but really it should work almost all the time anyway. */
    block=(((unsigned char) rbuff[2])*256) + (unsigned char) rbuff[3];
    if (block!=curblock+1) {
      if (DEBUG) printf("Out of order block received, not going to ack it\n");
      continue;
    }
    curblock++;
    make_ack(sbuff, block);
    sendto(sockfd, sbuff, 4, NULL, (struct sockaddr *) &cliaddr, sizeof(cliaddr));
    if (DEBUG) printf("Received data packet for block %i, len %i, and sent ack\n", block, packetlen-4);

    /* bail if it's a packet with no data (a mod 512 byte file that finished last packet) */
    if (ret==4) {
      return(0);
    }

    /* translate netascii to ascii, get the length of the new buffer in len */
    if (mode==MODE_NETASCII) {
      len=un_netascii(rbuff+4, packetlen-4);
    } else {
      len=packetlen-4;
    }

    /* append to the file */
    write(fileno(file), rbuff+4, len);

    /* if there's no more data then we're done */
    if (packetlen<(512+4)) {
      return(0);
    }
  }
}

void make_ack(char *buff, int block) {
  /* make a TFTP ack packet for the block number 'block' */
  buff[0]=0;
  buff[1]=4;
  buff[2]=block>>8;
  buff[3]=block&255;
}

int un_netascii(char *buff, int len) {
  /* convert netascii to ascii in buff and return the new length */
  /* for now we'll just strip \r's whicih should be good enough for
     our purposes */
  char tmpbuff[len];
  int i, j;

  j=0;
  for (i=0; i<len; i++) {
    if (buff[i]!='\r') {
      tmpbuff[j]=buff[i];
      j++;
    }
  }

  memcpy(buff, tmpbuff, len);
  return(j);
}
